package at.bestsolution.rsd.rt.client;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import org.eclipse.microprofile.config.ConfigProvider;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.osgi.service.component.annotations.ReferencePolicy;
import org.osgi.service.component.annotations.ReferencePolicyOption;

@Component
public class ServiceLookupRegistry {
	public static final String LOOKUP_ORDER_CONFIG_KEY = "rsd.service.lookup.order";
	
	private Map<String,List<ServiceLookup>> lookups = new HashMap<>();
	
	@Reference(cardinality = ReferenceCardinality.AT_LEAST_ONE, policy = ReferencePolicy.DYNAMIC, policyOption = ReferencePolicyOption.GREEDY)
	public void registerServiceLookup(ServiceLookup lookup, Map<String, Object> properties) {
		synchronized (lookups) {
			List<ServiceLookup> list = lookups.computeIfAbsent((String)properties.getOrDefault(ServiceLookup.SERVICE_TYPE, "no-key"), key -> new ArrayList<>());
			list.add(lookup);
		}
	}
	
	public void unregisterServiceLookup(ServiceLookup lookup, Map<String, Object> properties) {
		synchronized (lookups) {
			List<ServiceLookup> list = lookups.computeIfAbsent((String)properties.getOrDefault(ServiceLookup.SERVICE_TYPE, "no-key"), key -> new ArrayList<>());
			list.remove(lookup);
		}
	}
	
//	public Set<String> getServiceTypes() {
//		return Collections.unmodifiableSet(lookups.keySet());
//	}
	
	public ServiceLookup getServiceLookup() {
		return new ServiceLookupImpl();
	}
	
	private <S> Optional<S> getService(List<ServiceLookup> list, Class<S> clazz) {
		Optional<S> serviceOpt =  list.stream()
				.map( l -> l.lookupService(clazz) )
				.filter( Optional::isPresent )
				.map( Optional::get )
				.findFirst();
		return serviceOpt;
	}
	
	class ServiceLookupImpl implements ServiceLookup {

		@Override
		public <S> Optional<S> lookupService(Class<S> clazz) {
			String[] serviceTypes = ConfigProvider.getConfig().getOptionalValue(LOOKUP_ORDER_CONFIG_KEY, String[].class).orElse(new String[] { ServiceLookup.EJB_TYPE, ServiceLookup.REST_TYPE });
			
			for( String serviceType : serviceTypes ) {
				Optional<S> serviceOpt = getService(lookups.getOrDefault(serviceType, Collections.emptyList()), clazz);
				if( serviceOpt.isPresent() ) {
					return serviceOpt;
				}
			}
			
			// Before we give up check all lookups
			for( Entry<String, List<ServiceLookup>> e : lookups.entrySet() ) {
				Optional<S>  serviceOpt = getService(e.getValue(), clazz);
				if( serviceOpt.isPresent() ) {
					return serviceOpt;
				}
			}
			return Optional.empty();
		}
		
	}
}
